/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT (c) 2008 - 2013 SRON (R.M.van.Hees@sron.nl)

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License, version 2, as
   published by the Free Software Foundation.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, 
   Boston, MA  02111-1307, USA.

.IDENTifer   NADC_MCFS
.AUTHOR      R.M. van Hees
.KEYWORDS    Ife MCFS product (GPOD)
.LANGUAGE    ANSI C
.PURPOSE     store parameters of a MCFS product SCIA in a PostgreSQL database
.INPUT/OUTPUT
  call as   
            nadc_mcfs [options] <input-file>

.RETURNS     non-negative on success, negative on failure
.COMMENTS    None
.ENVIRONment None
.VERSION     1.1     20-Oct-2008   conform to ADAGUC filename format, RvH
             1.0     15-Jul-2008   initial release by R. M. van Hees
------------------------------------------------------------*/
/*
 * Define _ISOC99_SOURCE to indicate
 * that this is a ISO C99 program
 */
#define  _ISOC99_SOURCE

/*+++++ System headers +++++*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WITH_NC4
#include <hdf5.h>
#include <netcdf.h>
#endif
#ifdef _WITH_SQL
#include <libpq-fe.h>
#endif

/*+++++ Local Headers +++++*/
#include <defs_nadc.h>
#include <defs_mcfs.h>

/*+++++ Macros +++++*/
#define NADC_PARAMS "\n\t" \
  "Usage: nadc_mcfs [-sql] [-remove] [-replace] <flname>"

/*+++++ Global Variables +++++*/
	/* NONE */

/*+++++ Static Variables +++++*/
	/* NONE */

/*+++++++++++++++++++++++++ Static Functions +++++++++++++++++++++++*/
static
void DateTime2ADAGUC( const char *dateTime, char *timeStamp )
{
     const char *mnth_lst[] = { 
	  "FAIL", "Jan", "Feb", "Mar", "Apr", "May", 
	  "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

     char         cbuff[4];
     unsigned int imin, ihour, iday, imon, iyear;
     double       sec;
/*
 * decomposition of date and time string into numbers
 */
     (void) sscanf( dateTime, "%2u-%3s-%4u %2u:%2u:%6lf",
		    &iday, cbuff, &iyear, &ihour, &imin, &sec );

     for ( imon = 1; imon <= 12; imon++ )
	  if ( strcmp( cbuff, mnth_lst[imon] ) == 0 ) break;

     (void) sprintf( timeStamp, "%04u%02u%02uT%02u%02u%02.0f",
                     iyear, imon, iday, ihour, imin, sec );
}

/*+++++++++++++++++++++++++ Main Program or Function +++++++++++++++*/
int main( int argc, char *argv[] )
{
     const char prognm[] = "nadc_mcfs";

#ifdef _WITH_NC4
     bool flag_sql     = FALSE;
#else
     bool flag_sql     = TRUE;
#endif
#ifdef _KNMI_SQL
     bool flag_remove  = FALSE;
     bool flag_replace = FALSE;
#endif
     char flname[MAX_STRING_LENGTH];

     struct mcfs_hdr hdr;
     struct mcfs_rec *rec = NULL;
/*
 * check command-line parameters
 */
     if ( argc > 2 ) {
          register int nr = 1;
          do {
               if ( strncmp( argv[nr], "-sql", 4 ) == 0 )
                    flag_sql    = TRUE;
#ifdef _KNMI_SQL
               else if ( strncmp( argv[nr], "-remove", 7 ) == 0 )
                    flag_remove = TRUE;
	       else if ( strncmp( argv[nr], "-replace", 8 ) == 0 )
		    flag_replace = TRUE;
#endif
               else
                    NADC_GOTO_ERROR( prognm, NADC_ERR_PARAM, NADC_PARAMS );
          } while( ++nr < (argc-1) );
          (void) strlcpy( flname, argv[argc-1], MAX_STRING_LENGTH );
     } else if ( argc == 2 ) {
          (void) strlcpy( flname, argv[1], MAX_STRING_LENGTH );
     } else
          NADC_GOTO_ERROR( prognm, NADC_ERR_PARAM, NADC_PARAMS );
/*
 * read the MCFS product
 */
     (void) NADC_RD_MCFS( flname, &hdr, &rec );
     if ( IS_ERR_STAT_FATAL )
          NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, "failed to read product" );
/*
 * connect to PostgreSQL database
 */
     if ( flag_sql ) {
#ifdef _KNMI_SQL
          PGconn *conn;

          NADC_CONNECT_DB( &conn, "scia" );
          if ( IS_ERR_STAT_FATAL ) {
               NADC_ERROR( prognm, NADC_ERR_SQL, "MCFS (connect)" );
               NADC_Err_Trace( stderr );
               return NADC_ERR_FATAL;
          }
	  if ( flag_remove || flag_replace ) {
	       NADC_MCFS_DEL_ENTRY( conn, hdr.product );
	       if ( IS_ERR_STAT_FATAL )
		    NADC_GOTO_ERROR( prognm, NADC_ERR_SQL, "MCFS (remove)" );
	  }
/*
 * write meta-information to database
 */
          if ( ! flag_remove ) {
               NADC_MCFS_WR_SQL_META( conn, &hdr );
               if ( IS_ERR_STAT_FATAL ) {
                    NADC_ERROR( prognm, NADC_ERR_SQL, "MCFS (header)" );
               } else {
                    NADC_MCFS_WR_SQL_TILE(conn, hdr.product, hdr.numRec, rec);
                    if ( IS_ERR_STAT_FATAL )
                         NADC_ERROR( prognm, NADC_ERR_SQL, "MCFS (records)" );
               }
          }
/*
 * close connection to PostgreSQL database
 */
          if ( conn != NULL ) PQfinish( conn );
#endif
     } else {
#ifdef _WITH_NC4
          char validityStart[16], validityStop[16];

          int  ncid;
          int  retval;

          /* ToDo: need to check that input/output file are not the same */
          DateTime2ADAGUC( hdr.product_datetime_start, validityStart );
          DateTime2ADAGUC( hdr.product_datetime_end, validityStop );

          (void) snprintf( flname, MAX_STRING_LENGTH, ADAGUC_PROD_TEMPLATE,
                           validityStart, validityStop, 1 );

          /* Create the file. The NC_NETCDF4 parameter tells netCDF to create
           * a file in netCDF-4/HDF5 standard.
           */
          if ( (retval = nc_create( flname, NC_NETCDF4, &ncid ))!= NC_NOERR )
               NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );

          NADC_MCFS_WR_NC_META( ncid, &hdr );
	  if ( IS_ERR_STAT_FATAL )
	       NADC_GOTO_ERROR( prognm, NADC_ERR_HDF_WR, 
				"problems with header" );
          NADC_MCFS_WR_NC_REC( ncid, hdr.numRec, rec );
	  if ( IS_ERR_STAT_FATAL )
	       NADC_GOTO_ERROR( prognm, NADC_ERR_HDF_WR, 
				"problems with data records" );

          if ( nc_close( ncid ) != NC_NOERR )
               NADC_GOTO_ERROR( prognm, NADC_ERR_FATAL, nc_strerror(retval) );
#endif
     }
/*
 * close input file
 */
 done:
     if ( rec != NULL ) free( rec );
/*
 * display error messages
 */
     NADC_Err_Trace( stderr );
     if ( IS_ERR_STAT_FATAL )
          return NADC_ERR_FATAL;
     else
          return NADC_ERR_NONE;
}
