/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT (c) 1999 - 2013 SRON (R.M.van.Hees@sron.nl)

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License, version 2, as
   published by the Free Software Foundation.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, 
   Boston, MA  02111-1307, USA.

.IDENTifer   defs_scia
.AUTHOR      R.M. van Hees 
.KEYWORDS    header file
.LANGUAGE    ANSI-C
.PURPOSE     macros and structures for SCIAMACHY data modules
.ENVIRONment none
.VERSION      4.2   06-Oct-2004	added coadd-factor to struct mds1c_scia, RvH
              4.1   31-Mar-2003	modified include for C++ code
              4.0   13-Sep-2001	separate 0/1b/2 specific definitions, RvH 
              3.0   29-Aug-2001 combines level 1b and 2 definitions, RvH
              2.0   04-Jan-2001 Updated to format: DDT v2.3+3 Ev 2.0, RvH
              1.0   12-Aug-1999 Creation by R.M. van Hees
------------------------------------------------------------*/
#ifndef  __DEFS_SCIA                            /* Avoid redefinitions */
#define  __DEFS_SCIA

#ifdef __cplusplus
extern "C" {
#endif

/*+++++ Macros +++++*/

/* maximum number of cluster definitions (?) */
#define MAX_CLUSTER    ((unsigned short) 64)

/* value for mds_flag when a MDS is attached */
#define MDS_ATTACHED   ((unsigned char) 0)

/* general Sciamachy Instrument definitions */
#define SCIENCE_CHANNELS    8
#define VIS_CHANNELS        5
#define IR_CHANNELS         3
#define FirstInfraChan ((unsigned char) (SCIENCE_CHANNELS-IR_CHANNELS+1))

#define PMD_NUMBER          7
#define IR_PMD_NUMBER       2

#define CHANNEL_SIZE        1024
#define ALL_CHANNELS        (SCIENCE_CHANNELS + PMD_NUMBER)
#define SCIENCE_PIXELS      (SCIENCE_CHANNELS * CHANNEL_SIZE)
#define VIS_SCIENCE_PIXELS  (VIS_CHANNELS * CHANNEL_SIZE)
#define IR_SCIENCE_PIXELS   (IR_CHANNELS * CHANNEL_SIZE)

#define NUM_FRAC_POLV       12
#define NUM_SPEC_COEFFS     5

/*+++++ Structures & Unions +++++*/
enum scia_state { SCIA_LVL0 = 0, 
		  SCIA_NADIR, SCIA_LIMB, SCIA_OCCULT, SCIA_MONITOR 
};

enum scia_polV { PMD_THEORY = 0,
		 PMD_1, PMD_2, PMD_3, PMD_4, PMD_5, PMD_6, 
		 PMD_OVL1, PMD_OVL2, PMD_OVL3, PMD_OVL4, PMD_OVL5 
};

enum pixel_type_nadir { BACK_SCAN = 0, FORWARD_SCAN, NO_PIXEL_TYPE = 255 
};

enum scia_q_sost { SCIA_Q_OK = 0,
		   SCIA_Q_DECON, SCIA_Q_RECOVER, SCIA_Q_UNAVAIL, 
		   SCIA_Q_UNKOWN = 255
};

#define ALONG_TANG_HGHT ((unsigned char) 0x0U)
#define NEW_TANG_HGHT   ((unsigned char) 0x1U)
#define DEEP_SPACE      ((unsigned char) 0x2U)

/*+++++ Global Variables +++++*/
/* 
 * Most routines to read SCIAMACHY data can allocate memory internally
 * However IDL requires the use of their own memory allocation routines
 */
extern bool Use_Extern_Alloc;

/*+++++ Structures & Unions +++++*/
/*
 * basic SCIAMACHY data structures
 */
struct geoL_scia
{
     unsigned char pixel_type;
     unsigned char glint_flag;
     float pos_esm;
     float pos_asm;
     float sat_h;
     float earth_rad;
     float dopp_shift;
     float sun_zen_ang[3];
     float sun_azi_ang[3];
     float los_zen_ang[3];
     float los_azi_ang[3];
     float tan_h[3];
     struct coord_envi sub_sat_point;
     struct coord_envi tang_ground_point[3];
};

struct geoN_scia
{
     unsigned char pixel_type;
     unsigned char glint_flag;
     float pos_esm;
     float sat_h;
     float earth_rad;
     float sun_zen_ang[3];
     float sun_azi_ang[3];
     float los_zen_ang[3];
     float los_azi_ang[3];
     struct coord_envi sub_sat_point;
     struct coord_envi corner[NUM_CORNERS];
     struct coord_envi center;
};

struct geoC_scia
{
     float pos_esm;
     float pos_asm;
     float sun_zen_ang;
     struct coord_envi sub_sat_point;
};

struct gdf_para {
     float p_bar;
     float beta;
     float w0;
};

struct polV_scia
{
     float Q[NUM_FRAC_POLV];
     float error_Q[NUM_FRAC_POLV];
     float U[NUM_FRAC_POLV];
     float error_U[NUM_FRAC_POLV];
     float rep_wv[NUM_FRAC_POLV+1];
     struct gdf_para gdf;
     unsigned short intg_time;
};

/* common SCIA PDS data structures */
struct lads_scia
{
     struct mjd_envi mjd;
     unsigned char flag_mds;
     struct coord_envi corner[NUM_CORNERS];
};

struct mds1c_scia
{
     struct mjd_envi  mjd;
     signed char      rad_units_flag;
     signed char      quality_flag;
     unsigned char    type_mds;
     unsigned char    coaddf;
     unsigned char    category;
     unsigned char    state_id;
     unsigned char    state_index;
     unsigned char    chan_id;
     unsigned char    clus_id;
     unsigned short   dur_scan;
     unsigned short   num_obs;
     unsigned short   num_pixels;
     unsigned int     dsr_length;
     float            orbit_phase;
     float            pet;
     unsigned short   *pixel_ids;     /* num_pixels */
     float            *pixel_wv;      /* num_pixels */
     float            *pixel_wv_err;  /* num_pixels */
     float            *pixel_val;     /* num_obs * num_pixels */
     float            *pixel_err;     /* num_obs * num_pixels */
     struct geoC_scia *geoC;          /* num_obs, Monitoring MDS only  */
     struct geoL_scia *geoL;          /* num_obs, Limb/Occultation MDS only  */
     struct geoN_scia *geoN;          /* num_obs, Nadir MDS only  */
};

struct mds1c_pmd
{
     struct mjd_envi  mjd;
     signed char      quality_flag;
     unsigned char    type_mds;
     unsigned char    category;
     unsigned char    state_id;
     unsigned char    state_index;
     unsigned short   dur_scan;
     unsigned short   num_pmd;
     unsigned short   num_geo;
     unsigned int     dsr_length;
     float            orbit_phase;
     float            *int_pmd;       /* num_pmd */
     struct geoL_scia *geoL;          /* num_obs, Limb/Occultation MDS only  */
     struct geoN_scia *geoN;          /* num_obs, Nadir MDS only  */
};

struct mds1c_polV
{
     struct mjd_envi  mjd;
     signed char      quality_flag;
     unsigned char    type_mds;
     unsigned char    category;
     unsigned char    state_id;
     unsigned char    state_index;
     unsigned short   dur_scan;
     unsigned short   total_polV;
     unsigned short   num_diff_intg;
     unsigned short   num_geo;
     unsigned int     dsr_length;
     float            orbit_phase;
     unsigned short   intg_times[MAX_CLUSTER];
     unsigned short   num_polar[MAX_CLUSTER];
     struct polV_scia *polV;          /* total_num_polV */
     struct geoL_scia *geoL;          /* num_obs, Limb/Occultation MDS only  */
     struct geoN_scia *geoN;          /* num_obs, Nadir MDS only  */
};

struct dmop_rec {
     char timeLine[8];
     char dateTimeStart[24];
     char dateTimeStop[24];
     unsigned int   muSecStart;
     unsigned int   muSecStop;
     float          orbitPhase;
     unsigned char  stateID;
     unsigned short absOrbit;
};

#ifdef _WITH_SQL
struct stateinfo_rec {
     char           softVersion[4];
     unsigned char  stateID;
     unsigned short indxState;
     unsigned int   indxDMOP;
     double         jday;
     double         dtMatch;
};

struct mds0_sql {
     struct mjd_envi mjd;
     unsigned char   stateID;
     unsigned short  nrAux;
     unsigned short  nrDet;
     unsigned short  nrPMD;
     float           obmTemp;
     float           pmdTemp;
     float           chanTemp[SCIENCE_CHANNELS];
};
#endif

/*
 * define scale/offset values for memory/non-linearity correction values
 */
#ifdef _MEMNLIN_CORR_OLD
static
const struct scale_rec nadc_scale[SCIENCE_CHANNELS] = {
     {-37.f, 1.25f}, {-37.f, 1.25f}, {-37.f, 1.25f}, {-37.f, 1.25f},
     {-37.f, 1.25f}, {-102.f, 1.25f}, {126.f, 1.5f}, {126.f, 1.25f}
};

static
const struct scale_rec atbd_scale[SCIENCE_CHANNELS] = {
     {0.f, 2.f}, {0.f, 2.f}, {0.f, 2.f}, {0.f, 2.f}, {0.f, 2.f},
     {-102.f, 1.25f}, {126.f, 1.5f}, {126.f, 1.25f}
};
#endif

#if (defined _SCIA_LEVEL_0)
#include <defs_scia_lv0.h>
#endif

#if (defined _SCIA_LEVEL_1) || (defined _SCIA_PATCH_1)
#include <defs_scia_lv0.h>
#include <defs_scia_lv1.h>
#endif
#ifdef _SCIA_LEVEL_2
#include <defs_scia_lv2.h>
#endif

#ifdef __cplusplus
  }
#endif
#endif /* __DEFS_SCIA */
