/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT (c) 2007 - 2013 SRON (R.M.van.Hees@sron.nl)

   This is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License, version 2, as
   published by the Free Software Foundation.

   The software is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, 
   Boston, MA  02111-1307, USA.

.IDENTifer   SCIA_LV1_DEL_ENTRY
.AUTHOR      R.M. van Hees
.KEYWORDS    SCIA SQL
.LANGUAGE    ANSI C
.PURPOSE     remove entries in database for a Sciamachy level 1b product
.INPUT/OUTPUT
  call as   SCIA_LV1_DEL_ENTRY( conn, be_verbose, flname );
     input:  
             PGconn *conn        :  PostgreSQL connection handle
	     bool be_verbose     :  be verbose
             char *flname        :  filename of the Sciamachy Level 1b product

.RETURNS     Nothing, error status passed by global variable ``nadc_stat''
.COMMENTS    Removes entries related to file "flname" from tables:
             meta__1P, stateinfo_meta__1P and reverse stateinfo
.ENVIRONment None
.VERSION     2.1     04-Oct-2012   added verbose flag, RvH
             2.0     08-Jun-2007   port to PostgreSQL R. M. van Hees
             1.0     06-Mar-2007   initial release by R. M. van Hees
------------------------------------------------------------*/
/*
 * Define _ISOC99_SOURCE to indicate
 * that this is a ISO C99 program
 */
#define  _GNU_SOURCE

/*+++++ System headers +++++*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <libpq-fe.h>

/*+++++ Local Headers +++++*/
#define _SCIA_LEVEL_1
#include <defs_nadc.h>

/*+++++ Macros +++++*/
#define SQL_STR_SIZE   256

#define SELECT_FROM_STATEINFO \
"SELECT unnest(fk_stateinfo) FROM stateinfo_meta__1P WHERE\
 fk_meta=(SELECT pk_meta FROM meta__1P where name=\'%s\')"

#define DELETE_FROM_META \
"DELETE FROM meta__1P WHERE name=\'%s\'"

#define UPDATE_STATEINFO \
"UPDATE stateinfo \
SET softVersion=SUBSTR(softVersion,1,1) || \'0\' || SUBSTR(softVersion,3,1) \
WHERE pk_stateinfo=%u AND SUBSTR(softVersion,2,1)=\'%s\'"

/*+++++ Global Variables +++++*/
	/* NONE */

/*+++++ Static Variables +++++*/
	/* NONE */

/*+++++++++++++++++++++++++ Static Functions +++++++++++++++++++++++*/
				/* NONE */

/*+++++++++++++++++++++++++ Main Program or Function +++++++++++++++*/
void SCIA_LV1_DEL_ENTRY( PGconn *conn, bool be_verbose, const char *flname )
{
     const char prognm[] = "SCIA_LV1_DEL_ENTRY";

     register int nr;

     char sql_query[SQL_STR_SIZE];

     char procStage[2];

     char         *pntr;
     int          nrow, numChar;
     unsigned int indx;

     PGresult *res, *res_update;
/*
 * check if we have to reverse field "softVersion[1]" in table "stateinfo"
 */
     numChar = snprintf( sql_query, SQL_STR_SIZE, SELECT_FROM_STATEINFO,
			 basename( flname ) );
     if ( be_verbose )
	  (void) printf( "%s(): %s [%-d]\n", prognm, sql_query, numChar );
     if ( numChar >= SQL_STR_SIZE )
	  NADC_RETURN_ERROR( prognm, NADC_ERR_STRLEN, "sql_query" );
     res = PQexec( conn, sql_query );
     if ( PQresultStatus( res ) != PGRES_TUPLES_OK ) {
	  NADC_RETURN_ERROR( prognm, NADC_ERR_SQL, PQresultErrorMessage(res) );
     }
/* return when no entries are found */
     if ( (nrow = PQntuples( res )) == 0 ) goto done;
/*
 * update field "softVersion[1]" in table "stateinfo"
 */
     (void) strlcpy( procStage, basename( flname )+10, 2 );
     for ( nr = 0; nr < nrow; nr++ ) {
	  pntr = PQgetvalue( res, nr, 0 );
          indx = (unsigned int) strtoul( pntr, (char **)NULL, 10 );
	  numChar = snprintf( sql_query, SQL_STR_SIZE, 
			      UPDATE_STATEINFO, indx, procStage );
	  if ( be_verbose )
	       (void) printf( "%s(): %s [%-d]\n", prognm, sql_query, numChar );
	  if ( numChar >= SQL_STR_SIZE )
	       NADC_RETURN_ERROR( prognm, NADC_ERR_STRLEN, "sql_query" );
	  res_update = PQexec( conn, sql_query );
	  if ( PQresultStatus( res_update ) != PGRES_COMMAND_OK ) {
	       NADC_ERROR( prognm, NADC_ERR_SQL, 
			   PQresultErrorMessage(res_update) );
	       PQclear( res_update );
	       goto done;
	  }
	  PQclear( res_update );
     }
 done:
     PQclear( res );
/*
 * remove entry from table "meta__1P"
 * referenced entries in table "stateinfo_meta__0P" are removed by the engine
 */
     numChar = snprintf( sql_query, SQL_STR_SIZE, DELETE_FROM_META,
			 basename( flname ) );
     if ( be_verbose )
	  (void) printf( "%s(): %s [%-d]\n", prognm, sql_query, numChar );
     if ( numChar >= SQL_STR_SIZE )
	  NADC_RETURN_ERROR( prognm, NADC_ERR_STRLEN, "sql_query" );
     res = PQexec( conn, sql_query );
     if ( PQresultStatus( res ) != PGRES_COMMAND_OK )
	  NADC_ERROR( prognm, NADC_ERR_SQL, PQresultErrorMessage(res) );
     PQclear( res );
}
